/*
 *  Internal Threads
 *
 *
 *  COPYRIGHT (c) 1989, 1990, 1991, 1992, 1993, 1994.
 *  On-Line Applications Research Corporation (OAR).
 *  All rights assigned to U.S. Government, 1994.
 *
 *  This material may be reproduced by or for the U.S. Government pursuant
 *  to the copyright license under the clause at DFARS 252.227-7013.  This
 *  notice must appear in all copies of this file and its derivatives.
 *
 *  intthrd.c,v 1.16 1995/12/19 20:02:47 joel Exp
 */

#include <rtems/system.h>
#include <rtems/score/chain.h>
#include <rtems/score/interr.h>
#include <rtems/intthrd.h>
#include <rtems/score/mpci.h>
#include <rtems/score/object.h>
#include <rtems/score/states.h>
#include <rtems/score/thread.h>
#include <rtems/score/userext.h>
#include <rtems/score/wkspace.h>

#include <rtems/io.h>
#include <rtems/score/sysstate.h>
#include <rtems/score/apiext.h>

/*PAGE
 *
 *  _Internal_threads_Start
 *
 *  This kernel routine initializes the internal thread handler.
 */

void _Internal_threads_Initialization( void )
{
}

/*PAGE
 *
 *  _Internal_threads_Start
 *
 *  This kernel routine creates and starts the system initialization and
 *  idle threads.  The priorities for the two threads must be adjusted to
 *  priority 0 (an illegal priority) to insure that no application thread
 *  runs until system initialization is complete.
 *
 *  Input parameters:   NONE
 *
 *  Output parameters:  NONE
 */

char *_Internal_thread_Idle_name = "IDLE";
char *_Internal_thread_SYSI_name = "SYSI";

void _Internal_threads_Start( void )
{
  void *idle;

  /*
   *  Register the MP Process Packet routine.
   */

  _MPCI_Register_packet_processor(
    MP_PACKET_INTERNAL_THREADS,
    _Internal_threads_MP_Process_packet
  );

  /*
   *  Initialize this class of objects.
   */ 

  _Objects_Initialize_information(
    &_Internal_threads_Information,
    OBJECTS_INTERNAL_THREADS,
    FALSE,
    2,
    sizeof( Thread_Control ),
    TRUE,
    8,
    TRUE
  );

  _Internal_threads_System_initialization_thread = _Internal_threads_Allocate();
  _Internal_threads_Idle_thread = _Internal_threads_Allocate(); 

  /*
   *  The entire workspace is zeroed during its initialization.  Thus, all
   *  fields not explicitly assigned were explicitly zeroed by
   *  _Workspace_Initialization.
   */

  _Thread_Initialize(
    &_Internal_threads_Information,
    _Internal_threads_System_initialization_thread,
    NULL,        /* allocate the stack */
    INTERNAL_THREADS_SYSTEM_INITIALIZATION_THREAD_STACK_SIZE,
    CPU_ALL_TASKS_ARE_FP,
    PRIORITY_MINIMUM,
    FALSE,       /* no preempt */
    FALSE,       /* not timesliced */
    0,           /* all interrupts enabled */
    _Internal_thread_SYSI_name
  );
 
  /*
   *  WARNING!!! This is necessary to "kick" start the system.
   */
  _Thread_Executing = _Internal_threads_System_initialization_thread;

  _Thread_Start(
    _Internal_threads_System_initialization_thread,
    THREAD_START_NUMERIC,
    _Internal_threads_System_initialization_thread_body,
    NULL,
    0
  );

  /*
   *  Initialize the IDLE task.
   */

#if (CPU_PROVIDES_IDLE_THREAD_BODY == TRUE)
  idle = _CPU_Internal_threads_Idle_thread_body;
#else
  idle = _Internal_threads_Idle_thread_body;
#endif

  if ( _CPU_Table.idle_task )
    idle = _CPU_Table.idle_task;

  _Thread_Initialize(
    &_Internal_threads_Information,
    _Internal_threads_Idle_thread,
    NULL,        /* allocate the stack */
    INTERNAL_THREADS_IDLE_THREAD_STACK_SIZE,
    CPU_IDLE_TASK_IS_FP,
    PRIORITY_MAXIMUM,
    TRUE,        /* preemptable */
    FALSE,       /* not timesliced */
    0,           /* all interrupts enabled */
    _Internal_thread_Idle_name
  );
 
  _Thread_Start(
    _Internal_threads_Idle_thread,
    THREAD_START_NUMERIC,
    idle,
    NULL,
    0
  );

}

/*PAGE
 *
 *  _Internal_threads_Idle_thread_body
 *
 *  This kernel routine is the idle thread.  The idle thread runs any time
 *  no other thread is ready to run.  This thread loops forever with
 *  interrupts enabled.
 *
 *  Input parameters:
 *    ignored - this parameter is ignored
 *
 *  Output parameters:  NONE
 */

#if (CPU_PROVIDES_IDLE_THREAD_BODY == FALSE)
Thread _Internal_threads_Idle_thread_body(
  unsigned32 ignored
)
{
  for( ; ; ) ;
}
#endif

/*PAGE
 *
 *  _Internal_threads_System_initialization_thread_body
 *
 *  This kernel routine is the system initialization thread.
 *  It is the first thread to execute because it performs
 *  system initialization functions such as calling
 *  _IO_initialization for each of the application's drivers.
 *  After initialization if multiprocessing is enabled the the thread
 *  becomes the multiprocessing receive thread, otherwise
 *  (i.e. in a single processor system) the thread kills itself.
 *
 *  Input parameters:
 *    ignored - ignores the parameter all threads receive
 *
 *  Output parameters:  NONE
 */

Thread _Internal_threads_System_initialization_thread_body(
  unsigned32 ignored
)
{
  /*
   *  Run the API and BSPs predriver hook.
   */

  _API_extensions_Run_predriver();
 
  if ( _CPU_Table.predriver_hook )
    (*_CPU_Table.predriver_hook)();

  /*
   *  Initialize all the device drivers and initialize the MPCI layer.
   *
   *  NOTE:  The MPCI may be build upon a device driver.
   */

  _IO_Initialize_all_drivers();

  if ( _System_state_Is_multiprocessing ) {
    _MPCI_Initialization();
    _Internal_threads_MP_Send_process_packet(
      INTERNAL_THREADS_MP_SYSTEM_VERIFY
    );
  }

  /*
   *  Run the APIs and BSPs postdriver hooks.
   */

  _API_extensions_Run_postdriver();
 
  if ( _CPU_Table.postdriver_hook )
    (*_CPU_Table.postdriver_hook)();


  /*
   *  The system is now multitasking and completely initialized.  
   *  This system thread now either "goes away" in a single processor 
   *  system or "turns into" the server thread in an MP system.
   */

  _System_state_Set( SYSTEM_STATE_UP );

  if ( _System_state_Is_multiprocessing )
    _MPCI_Receive_server();                            /* DOES NOT RETURN */

  /*
   *  Conceptually this is a simple delete of self by the system intialization
   *  thread.   In reality all this does is permanently block the system
   *  initialization thread.  The stack is freed while the thread control 
   *  block for this thread is not freed.  There is no way for this 
   *  thread to ever run again.
   *
   *  NOTE: This code is only executed in a single processor configuration.
   */

  _Thread_Disable_dispatch();
  _Thread_Close( &_Internal_threads_Information, _Thread_Executing );
  _Thread_Enable_dispatch();
}
