/*
 *-------------------------------------------------------------------
 *
 *   tpuInline.h -- inline functions for tpuInline.c
 *
 * This file has been created by John S. Gwynne for support of 
 * Motorola's 68332 MCU in the efi332 project.
 * 
 * Redistribution and use in source and binary forms are permitted
 * provided that the following conditions are met:
 * 1. Redistribution of source code and documentation must retain
 *    the above authorship, this list of conditions and the
 *    following disclaimer.
 * 2. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * This software is provided "AS IS" without warranty of any kind,
 * either expressed or implied, including, but not limited to, the
 * implied warranties of merchantability, title and fitness for a
 * particular purpose.
 *
 *------------------------------------------------------------------
 */

#ifndef _tpuInline_h_
#define _tpuInline_h_

#include "stdlib.h"
#include "tpu.h"
#include "tpudriver.h"

static void tpuStopAllChannels(void);
__inline__ static void tpuStopAllChannels(void) {
#if TPUincludeSimualtor == 0
    *CPR0 = *CPR1 = 0;
#else
    if (TPUsimulator<8) {
	*CPR1 &= 0x3<<(TPUsimulator*2);
	*CPR0 = 0;
    } else {
	*CPR1 = 0;
	/* abs() just stops a compilation warning */
	*CPR0 &= 0x3<<(abs(TPUsimulator-8)*2);
    }
#endif
}

static void tpuLoadEmulationRAM(void);
__inline__ static void tpuLoadEmulationRAM(void) {
    /* Configure the TRAM */
    *TRAMBAR = ((TRAM_START>>8) & TRAM_ADDR); /* Base address 0xf0000 */
    *TRAMMCR = 0;
    /* upload the micro code */
    bcopy(tpu_buffer,(unsigned char *)TRAM_START,TPU_RAM_SIZE);
}

static void tpuInitTPUMCRandTICR(void);
__inline__ static void tpuInitTPUMCRandTICR(void) {
    *TPUMCR = ( SAM(TPUTCR1,13,TPU_TCR1P) | /* TCR1 Prescaler to 1 */
		SAM(EFI_TPU_IARB,0,TPU_IARB) | /* Interrupt Arbitration */
		TPU_EMU | SAM(TPUPSCK,6,TPU_PSCK));
    *TICR = (SAM(EFI_TPU_CIRL,8,TPU_CIRL) | SAM(EFI_TPU_CIBV,4,TPU_CIBV));
}

static void tpuInterruptCatch(rtems_isr_entry ISR, rtems_vector_number vect);
__inline__ static void tpuInterruptCatch(rtems_isr_entry ISR, 
					 rtems_vector_number vect) {
    rtems_isr_entry dummy;
    rtems_status_code status;

    status = rtems_interrupt_catch(ISR,vect,&dummy);
    directive_failed( status, "rtems_interrupt_catch");
}

static void tpuInstallInterruptVectors(void);
__inline__ static void tpuInstallInterruptVectors(void) {
    int i;
    tpuInterruptCatch(crankISR, TPUcrank+EFI_TPU0);
    for (i=0;i<numberInjectors;i++)
	tpuInterruptCatch(injectorISR,TPUinjector[i]+EFI_TPU0);
    for (i=0;i<numberIgnition;i++)
	tpuInterruptCatch(ignitionISR,TPUignition[i]+EFI_TPU0);
    if (TPUincludeSimualtor)
	tpuInterruptCatch(simulatorISR,TPUsimulator+EFI_TPU0);
}

static void tpuSetFunctionSelectRegisters(int channel, int function);
__inline__ static void tpuSetFunctionSelectRegisters(int channel, int function)
{
    volatile unsigned short int *loc = CFSR0 + (3-channel/4);
    channel %= 4;
    *loc = (*loc & ~(0xf<<(channel*4))) | function<<(channel*4); 
}

static void tpuInitFunctionSelectRegisters();
__inline__ static void tpuInitFunctionSelectRegisters() {
    int i;
    tpuSetFunctionSelectRegisters(TPUcrank,0xf);
    for (i=0;i<numberInjectors;i++)
	tpuSetFunctionSelectRegisters(TPUinjector[i],0xe);
    for (i=0;i<numberIgnition;i++)
	tpuSetFunctionSelectRegisters(TPUignition[i],0xe);
    if (TPUincludeSimualtor)
	tpuSetFunctionSelectRegisters(TPUsimulator,0xc);
}

static void tpuSetHostSequenceRegisters(int channel, int value);
__inline__ static void tpuSetHostSequenceRegisters(int channel, int value) {
    volatile unsigned short int *loc = HSQR0 + (1-channel/8);
    channel %= 8;
    *loc = (*loc & ~(0x3<<(channel*2))) | value<<(channel*2); 
}

static void tpuInitHostSequenceRegisters();
__inline__ static void tpuInitHostSequenceRegisters(){
    int i;
    
    for (i=0;i<numberInjectors;i++)
	tpuSetHostSequenceRegisters(TPUinjector[i],0x01);
    for (i=0;i<numberIgnition;i++)
	tpuSetHostSequenceRegisters(TPUignition[i],0x00);
}

static void tpuSetChannelInterruptRegister(int channel, int value);
__inline__ static void tpuSetChannelInterruptRegister(int channel, int value){
    *CIER = (*CIER & ~(1<<channel)) | value<<channel; 
}

static void tpuInitChannelInterruptRegister();
__inline__ static void tpuInitChannelInterruptRegister(){
    int i;
    tpuSetChannelInterruptRegister(TPUcrank,1);
    for (i=0;i<numberInjectors;i++)
	tpuSetChannelInterruptRegister(TPUinjector[i],1);
    for (i=0;i<numberIgnition;i++)
	tpuSetChannelInterruptRegister(TPUignition[i],1);
    if (TPUincludeSimualtor)
	tpuSetChannelInterruptRegister(TPUsimulator,1);
}

static void tpuLoadSimPRAM();
__inline__ static void tpuLoadSimPRAM(){
    *(TPU_PR0+TPUsimulator*8+0) = 
       (SAM(0x4,5,0x1e0) | SAM(0x4,2,0x1c) | SAM(0x1,0,0x3));
    *(TPU_PR0+TPUsimulator*8+2) = tpuSimPeriod/2;
    *(TPU_PR0+TPUsimulator*8+3) = tpuSimPeriod;
}

static void tpuLoadCrankPRAM();
__inline__ static void tpuLoadCrankPRAM(){
    *(TPU_PR0+TPUcrank*8+0) = PMMXconfig;           
    *(TPU_PR0+TPUcrank*8+1) = PMMXwaitTeeth;
    *(TPU_PR0+TPUcrank*8+2) = PMMXintTooth;
    *(TPU_PR0+TPUcrank*8+3) = PMMXstartPeriod<<8 | PMMXnumberPeriods;
    *(TPU_PR0+TPUcrank*8+4) = PMMXratio<<8;
    *(TPU_PR0+TPUcrank*8+5) = PMMXlinkStart<<12 | (PMMXlinkCount<<8);
}

static void tpuLoadInjectorPRAM();
__inline__ static void tpuLoadInjectorPRAM(){
    int i;

#warning fix PRAM!!!!!!!!!!!!!!!!!
    for (i=0;i<numberInjectors;i++) {
	*(TPU_PR0+TPUinjector[i]*8+0) = 0;
	*(TPU_PR0+TPUinjector[i]*8+1) = 0;
	*(TPU_PR0+TPUinjector[i]*8+2) = 0;
	*(TPU_PR0+TPUinjector[i]*8+3) = 0;
	*(TPU_PR0+TPUinjector[i]*8+4) = 0;
	*(TPU_PR0+TPUinjector[i]*8+5) = 0;
    }
}

static void tpuLoadIgnitionPRAM();
__inline__ static void tpuLoadIgnitionPRAM(){
    int i;

#warning fix PRAM!!!!!!!!!!!!!!!!!
    for (i=0;i<numberIgnition;i++) {
	*(TPU_PR0+TPUignition[i]*8+0) = 0x8a;
	*(TPU_PR0+TPUignition[i]*8+1) = 1500;
	*(TPU_PR0+TPUignition[i]*8+2) = 0x0a<<8;
	*(TPU_PR0+TPUignition[i]*8+3) = 0;
	*(TPU_PR0+TPUignition[i]*8+4) = 0x0a<<8;
	*(TPU_PR0+TPUignition[i]*8+5) = 60<<8 | ((int)(TPU_PR1+4) & 0x7e);
    }
}

static void tpuStartChannel(int channel, int req, int priority);
__inline__ static void tpuStartChannel(int channel, int req, int priority){
    if (channel<8) {
	*HSRR1 = (*HSRR1 & ~(3<<(channel*2))) | req<<(channel*2);
	*CPR1  = (*CPR1  & ~(3<<(channel*2))) | priority<<(channel*2);
	while (*HSRR1 & (3<<(channel*2)));
    } else {
	*HSRR0 = (*HSRR0 & ~(3<<((channel-8)*2))) | req<<((channel-8)*2);
	*CPR0  = (*CPR0  & ~(3<<((channel-8)*2))) | priority<<((channel-8)*2);
	while (*HSRR0 & (3<<((channel-8)*2)));
    }
}

static void tpuStartCrank(void);
__inline__ static void tpuStartCrank(void){
    tpuStartChannel(TPUcrank,0x3,crankPriority);
}

static void tpuStartSimulator(void);
__inline__ static void tpuStartSimulator(void){
    tpuStartChannel(TPUsimulator,0x2,simulatorPriority);
}

static void tpuStartInjector(void);
__inline__ static void tpuStartInjector(void){
    int i;
    for (i=0;i<numberInjectors;i++)
	tpuStartChannel(TPUinjector[i],0x3,injectorPriority);
}

static void tpuStartIgnition(void);
__inline__ static void tpuStartIgnition(void){
    int i;
    for (i=0;i<numberIgnition;i++)
	tpuStartChannel(TPUignition[i],0x3,ignitionPriority);
}

static void tpuStopChannel(int channel);
__inline__ static void tpuStopChannel(int channel){
    if (channel<8)
	*CPR1  &= ~(0x3<<(channel*2));
    else
	*CPR0  &= ~(0x3<<((channel-8)*2));
}

static void tpuForceChannelLow(int channel, int priority);
__inline__ static void tpuForceChannelLow(int channel, int priority) {
    *(TPU_PR0+channel*8+0) = 0x13;
    *(TPU_PR0+channel*8+1) = 0x00;
    tpuStartChannel(channel,0x3,priority);
}

static void tpuForceChannelsLow(void);
__inline__ static void tpuForceChannelsLow(void) {
    int i;
    for (i=0;i<numberIgnition;i++)
	tpuForceChannelLow(TPUignition[i],ignitionPriority);
    for (i=0;i<numberInjectors;i++)
	tpuForceChannelLow(TPUinjector[i],injectorPriority);
}

static void tpuForceInjectorsLow(void);
__inline__ static void tpuForceInjectorsLow(void) {
    int i;
    for (i=0;i<numberInjectors;i++)
	tpuForceChannelLow(TPUinjector[i],injectorPriority);
}

static void tpuIntTooth(int count);
__inline__ static void tpuIntTooth(int count) {
    /* not able to really write bytes to PRAM ??? */
    *((unsigned char*)(TPU_PR0)+TPUcrank*16+5)=(unsigned char)count;
    /* zero the status number.... again :( */
    *(TPU_PR0+TPUcrank*8+2) &= 0xff;
}

static void tpuIgnShutDownDelay(int sleepTime);
__inline__ static void tpuIgnShutDownDelay(int sleepTime) {
    rtems_status_code status;
    if ((sleepTime = (sleepTime*TICKS_PER_SECOND) / 1000) == 0) sleepTime=1;
    status = rtems_task_wake_after(sleepTime);
    directive_failed( status, "rtems_task_wake_after" );
}

static void tpuFlushMessageQueue(void);
__inline__ static void tpuFlushMessageQueue(void) {
    rtems_status_code status;
    int count;
    status = rtems_message_queue_flush( Queue_id[TPUQueueID], &count );
    directive_failed( status, "rtems_message_queue_flush" );
}

static void tpuClearInterrupt(int channel);
static void tpuClearInterrupt(int channel) {
        *CISR &= ~(0x01<<channel);
}

static int tpuCrankStat(void);
__inline__ static int tpuCrankStat(void) {
    int i;
    i = (int)( *(volatile unsigned char * const)(TPU_PR0+TPUcrank*8+2));
    *(TPU_PR0+TPUcrank*8+2) &= 0xff;
    return i;
}

static void tpuReadTPUqueue(void *argument);
__inline__ static void tpuReadTPUqueue(void *argument) {
    rtems_status_code status;
    int length;
    status = rtems_message_queue_receive(
       Queue_id[ TPUQueueID ],
       argument,
       &length,
       RTEMS_DEFAULT_OPTIONS,
       RTEMS_NO_TIMEOUT
       );
    directive_failed( status, "rtems_message_queue_receive" );
}

static void tpuDIOignHold(void);
__inline__ static void tpuDIOignHold(void){
    int i;
    for (i=0;i<numberIgnition;i++) {
	tpuStopChannel(TPUignition[i]);
	tpuSetChannelInterruptRegister(TPUignition[i],0);
	tpuSetFunctionSelectRegisters(TPUignition[i],0xb);
	if (*(TPU_PR0+TPUignition[i]*8+5) & 1)
	    tpuStartChannel(TPUignition[i],1,ignitionPriority);
	else
	    tpuStartChannel(TPUignition[i],2,ignitionPriority);
	tpuStopChannel(TPUignition[i]);
	tpuClearInterrupt(TPUignition[i]);
	tpuSetChannelInterruptRegister(TPUignition[i],1);
	tpuSetFunctionSelectRegisters(TPUignition[i],0xe);
    }
}


#endif /* _tpuInline_h_ */
